#include <iostream>
#include <filesystem>
#include <cmath>

//These are external libraries
#include <GLAD/glad.h> //for initializing opengl functions
#include <GLFW/glfw3.h> //for window context management

//for mathematical operations on vectors and matrices
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>

//audio library
#include <irrklang/irrKlang.h>

//for loading images
#define STB_IMAGE_IMPLEMENTATION
#include "stb_image.h"

//these are our own classes/framework
#include "shader.h"
#include "texture.h"
#include "camera.h"
#include "model.h"
#include "sphere.h"
#include "skybox.h"
#include "framebuffer.h"
#include "uniformbuffer.h"
#include "light.h"
#include "VolumetricFog.h"

//declaration of callback functions for interacting with the application, function definitions are found on the bottom of the file
void framebuffer_size_callback(GLFWwindow* window, int width, int height); //when the window gets resized
void mouse_callback(GLFWwindow* window, double xpos, double ypos); //when the mouse gets moved
void scroll_callback(GLFWwindow* window, double xoffset, double yoffset); //when the scroll wheel is used
void processInput(GLFWwindow* window); //for keyboard input

// settings
const int SCR_WIDTH = 1920;
const int SCR_HEIGHT = 1080;

const int ENVIRONMENT_MAP_DIMENSIONS = 512; //we use this as the dimensions for example when rendering shadowmaps (these have to be square)

//filepaths for shader files
const char* lightingVertexPath = "src/shader/lighting.vert";
const char* lightingFragmentPath = "src/shader/lighting.frag";
const char* lightSourceFragmentPath = "src/shader/lightSource.frag";
const char* tessControlPath = "src/shader/tess_control.tesc";
const char* tessEvalPath = "src/shader/tess_eval.tese";
const char* lightingWindVertexPath = "src/shader/lightingWind.vert";
const char* windGeoPath = "src/shader/wind.geo";

const char* reflectionFragmentPath = "src/shader/reflection.frag";

const char* skyboxVertexPath = "src/shader/skybox.vert";
const char* skyboxFragmentPath = "src/shader/skybox.frag";

const char* billboardVertexPath = "src/shader/billboard.vert";
const char* billboardFragmentPath = "src/shader/billboard.frag";

const char* quadVertexPath = "src/shader/quad.vert";
const char* quadFragmentPath = "src/shader/quad.frag";

const char* waterFragmentPath = "src/shader/water.frag";
const char* waterVertexPath = "src/shader/water.vert";

const char* depthFragmentPath = "src/shader/depth.frag";
const char* depthVertexPath = "src/shader/depth.vert";
const char* depthGeometryPath = "src/shader/depth.geo";

const char* environmentFragmentPath = "src/shader/environment.frag";
const char* environmentVertexPath = "src/shader/environment.vert";
const char* environmentGeometryPath = "src/shader/environment.geo";

const char* tangentSpaceVertexPath = "src/shader/tangentspace.vert";
const char* tangentSpaceFragmentPath = "src/shader/tangentspace.frag";
const char* tangentSpaceGeometryPath = "src/shader/tangentspace.geo";

const char* debugVertexPath = "src/shader/debug.vert";
const char* debugFragmentPath = "src/shader/debug.frag";
const char* debugGeometryPath = "src/shader/debug.geo";

const char* quadHDRFragmentPath = "src/shader/quadhdr.frag";

const char* textFragmentPath = "src/shader/text.frag";

// Compute shaders for volumetric fog
const char* lightInjectionPath = "src/shader/lightinjection.comp";
const char* rayMarchPath = "src/shader/raymarch.comp";

const char* camDepthVert = "src/shader/camDepth.vert";
const char* camDepthFrag = "src/shader/camDepth.frag";

//variable for mouse movement callback
bool firstMouse = true;

//Variables for key press events
bool bPressed = false;
bool cPressed = false;
bool fPressed = false;
bool mPressed = false;
bool oPressed = false;
bool rPressed = false;
bool tPressed = false;
bool uPressed = false;
bool vPressed = false;
bool plusPressed = false;
bool minusPressed = false;
bool kp6Pressed = false;
bool kp9Pressed = false;

bool k1Pressed = false;
bool k2Pressed = false;
bool k3Pressed = false;
bool k4Pressed = false;

//for calculation of deltatime between frames, for example to make camera movement speed independent from framerate
float startTime;
float timeDifference = 0.0f;
float startDifference;
float lastTime = 0.0f;
float turbo = 1.0f;
float cameraSpeed;

//initialize Camera
Camera camera;
//glm::vec3 initialCameraPosition = glm::vec3(0.0f, 10.0f, 5.0f);
glm::vec3 initialCameraPosition = glm::vec3(54.4733f, 16.427f, -20.8804f);
//glm::vec3 initialCameraPosition = glm::vec3(3.20711f, 2.5444f, -37.2465f);
bool userControlledCamera = false; // Switch between auto pathing and debug camera
std::vector<CameraTransform> cameraRecording;
bool recordCameraMovement = false; // When true, fill cameraRecording with camera positions and orientations

bool moonDisabled = false;

UserInput currentUserInput = {};
glm::vec4 anisotropy_density = glm::vec4(0.3f, 1.5f, 0.0f, 0.0f); // Volumetric Fog parameters, keep global (Changeable by keypress) 
glm::vec4 exposure_gamma = glm::vec4(2.0f, 0.7f, 0.0f, 0.0f); // Tone mapping parameters, keep global (Changeable by keypress) 

int main(void)
{
	// Initialize GLFW and create window
	if (!glfwInit())
		return -1;
	GLFWwindow* window;

	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 6);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
	window = glfwCreateWindow(SCR_WIDTH, SCR_HEIGHT, "RTR24", NULL, NULL);
	if (!window)
	{
		const char* description;
		int errorNum = glfwGetError(&description);
		std::cout << "Failed to create GLFW window" << std::endl;
		printf("Error %d: %s\n", errorNum, description);
		glfwTerminate();
		return -1;
	}
	glfwMakeContextCurrent(window);
	glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
	glfwSetFramebufferSizeCallback(window, framebuffer_size_callback);
	glfwSetCursorPosCallback(window, mouse_callback);
	glfwSetScrollCallback(window, scroll_callback);

	// Initialize GLAD (OpenGL functions bindings)
	if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
	{
		std::cout << "Failed to initialize GLAD" << std::endl;
		return -1;
	}
	printf("GL_VERSION: %s\n", glGetString(GL_VERSION));

	//initialize Sound Engine and load sounds
	irrklang::ISoundEngine* SoundEngine = irrklang::createIrrKlangDevice();
	if (!SoundEngine)
		std::cout << "ERROR::SOUND ENGINE FAILED TO START" << std::endl;
	//irrklang::ISoundSource* ambientSound = SoundEngine->addSoundSourceFromFile("sounds/DarkSoulsAmbience.mp3");
	//irrklang::ISoundSource* areaSound = SoundEngine->addSoundSourceFromFile("sounds/DarkSoulsNewArea.mp3");
	//ambientSound->setDefaultVolume(0.4f);
	//areaSound->setDefaultVolume(1.0f);

	//shaders
	Shader lightSourceShader(lightingVertexPath, lightSourceFragmentPath, NULL);
	Shader lightingShader(lightingVertexPath, lightingFragmentPath, NULL);
	Shader modelWindShader(lightingWindVertexPath, lightingFragmentPath, windGeoPath);
	Shader modelTessWindShader(lightingWindVertexPath, lightingFragmentPath, tessControlPath, tessEvalPath, windGeoPath);
	//Shader reflectionShader(lightingVertexPath, reflectionFragmentPath, NULL);
	Shader skyboxShader(skyboxVertexPath, skyboxFragmentPath, NULL);
	Shader billboardShader(billboardVertexPath, billboardFragmentPath, NULL);
	Shader quadShader(quadVertexPath, quadFragmentPath, NULL);
	Shader waterShader(waterVertexPath, waterFragmentPath, NULL);
	Shader depthMapShader(depthVertexPath, depthFragmentPath, depthGeometryPath);
	Shader depthTessShader(depthVertexPath, depthFragmentPath, tessControlPath, tessEvalPath, depthGeometryPath);
	Shader environmentShader(environmentVertexPath, environmentFragmentPath, environmentGeometryPath);
	Shader tangentSpaceShader(tangentSpaceVertexPath, tangentSpaceFragmentPath, tangentSpaceGeometryPath);
	Shader debugShader(debugVertexPath, debugFragmentPath, debugGeometryPath);
	Shader hdrShader(quadVertexPath, quadHDRFragmentPath, NULL);
	Shader textShader(quadVertexPath, textFragmentPath, NULL);

	Shader camDepthShader(camDepthVert, camDepthFrag, NULL);

	//models
	//The Sphere class implements a sphere rendered on a square billboard
	//Sphere sphere(glm::vec3(0.0f, 5.0f, 0.0f), 1.0f);
	Model quad("models/quad/quad.obj");
	Model lamp("models/light/light.obj");
	Model boat("models/ship/ship.obj");
	glm::vec3 boatPosition = glm::vec3(0,-0.75,0);
	Model moon("models/moon/moon.obj");
	Model island("models/island/island_low.obj");
	glm::vec3 islandPosition = glm::vec3(10, -1, 20);

	//for now the water is rendered using a simple plane that lies parallel to the xz-plane
	//this variable defines the y coordinate of the waterplane
	float waterlevel = -0.5;

	//textures
	Cubemap cm;
	cm.loadImages("textures/green_nebula/");
	Skybox skybox(cm);
	Texture2D waterDuDv;
	waterDuDv.loadImage("textures/water/wdudv.jpg");
	Texture2D waterNormal;
	waterNormal.loadImage("textures/water/normal.png");
	Texture2D areaName;
	//areaName.loadImage("textures/text/No Mans Galleon Bloodborne Style.png");
	areaName.loadImage("textures/text/intro_text.png");


	//lights
	//the shaders can handle a maximum of 11 lights including directional and point lights (at the moment directional lights are not fully supported)

	vector<Light> lights;
	vector<Light> lightsWithoutMoon;
	vector<Light> lightsWithMoon;

	glm::vec4 pierLampPosition = glm::vec4(4.9, 4.8, 3.25, 1.0);
	//glm::vec4 pierLampPosition = glm::vec4(4.9, 4.8, 10, 1.0);
	glm::vec3 pierLampColor = glm::vec3(0.9, 0.7, 0.1) * 0.5f;
	vector<glm::vec3> pierLampColorVec{ pierLampColor*0.1f,pierLampColor,pierLampColor };
	glm::vec3 pierLampAtt(1, 0.09/2, 0.032/2);
	lightsWithMoon.push_back(Light(pierLampColorVec, pierLampPosition, pierLampAtt));
	lightsWithoutMoon.push_back(Light(pierLampColorVec, pierLampPosition, pierLampAtt));
	
	//lights.push_back(Light(whiteVec, glm::vec4(10, 2, -10, 1), attenuation));
	// for the moon we render it at a fake position and not the position where the actual moonlight point light source is
	// the problem is that we modeled the moon as a point light source and not a directional light source
	//glm::vec4 moonLightPosition = glm::vec4(-20, 15, 60, 1);
	glm::vec4 moonLightPosition = glm::vec4(-15, 20, 60, 1);
	//float moonScale = 20.0f;
	float moonScale = 4.0f;
	glm::vec3 moonFakePosition = glm::vec3(-5, 9, 40) * moonScale;
	moonLightPosition = glm::vec4(moonFakePosition, 1.0);

	glm::vec3 moonLightColor = glm::vec3(1, 1, 1) * 0.05f;
	vector<glm::vec3> moonLightColorVec{ moonLightColor, moonLightColor, moonLightColor };
	glm::vec3 moonAtt(1, 0, 0);
	Light moonLight(moonLightColorVec, moonLightPosition, moonAtt);
	moonLight.far = 500.0f; // moon needs a bigger far plane as it is further aways
	lightsWithMoon.push_back(moonLight);


	lights = lightsWithMoon;
	//glm::vec3 blueLight = glm::vec3(0.5f, 0.7f, 0.9f);
	//lights.push_back(Light(vector<glm::vec3> { 0.1f * blueLight, blueLight, blueLight }, glm::vec4(0, 3.25, -6, 1), glm::vec3(1.0f, 0.5f, 0.2f)));

	//instead of manually passing uniform variables, we sometimes use Uniform Buffers for uniforms that get used by multiple shaders
	//lightUniforms is a Uniform buffer that contains the data for all the current lights
	LightUniformBuffer lightUniforms;
	lightUniforms.updateLightUniforms(lights);

	//declaration of matrices, that we will reuse often
	glm::mat4 model, view, projection, scale, rotate, translate;
	glm::mat3 normalMat;
	ViewProjectionClipUniformBuffer viewProjectionClipUniforms;

	UserInputUniformBuffer userInputUniforms;

	//framebuffers
	FramebufferHDR fbHDR(SCR_WIDTH, SCR_HEIGHT);
	Framebuffer waterReflectionFB(ENVIRONMENT_MAP_DIMENSIONS, ENVIRONMENT_MAP_DIMENSIONS);
	Framebuffer waterRefractionFB(ENVIRONMENT_MAP_DIMENSIONS, ENVIRONMENT_MAP_DIMENSIONS);

	//the depth framebuffer gets one cubemap array for color attachment and one cubemap array for depth attachment
	//each layer corresponds to one light source
	Framebuffer depthFB(ENVIRONMENT_MAP_DIMENSIONS, ENVIRONMENT_MAP_DIMENSIONS);
	depthFB.bind();
	EnvironmentMap3D envMaps(ENVIRONMENT_MAP_DIMENSIONS, lights.size());
	DepthCubemap3D depthMaps(ENVIRONMENT_MAP_DIMENSIONS, lights.size());
	glFramebufferTexture(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, depthMaps.ID, 0);
	glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, envMaps.ID, 0);
	
	// Initialization of the volumetric light
	VoxelGridDimension voxelGridDim{ 160,90,128 };
	VolumetricFog volFog(rayMarchPath, lightInjectionPath, voxelGridDim);

	camera.position = initialCameraPosition;
	camera.yaw = -223;
	camera.pitch = -10;
	//camera.yaw = -270.2f;
	//camera.pitch = 1.09999f;
	float farPlaneCamera = 1000.0f;
	float nearPlaneCamera = 0.1f;

	if (std::filesystem::exists("camPath.txt"))
		camera.readCameraPathFile("camPath.txt");
	if (!userControlledCamera)
		camera.startAutoPathing();

	glClearColor(1.0f, 0.0f, 0.0f, 1.0f);
	int firstTime = 1;
	int firstSound = 1;
	int lightMoveCnt = 0; // Variable for light movement

	currentUserInput.volFogAct_blueNoise_filterAct_moveLight	= glm::ivec4(1.0f, 1.0f, 1.0f, 0.0f);
	currentUserInput.windAct_tessEnable							= glm::ivec4(1.0f, 1.0f, 0.0f, 0.0f);

	float camTransformRecordingTime = 0.0f; // Time accumulation for camera recording
	float SAVE_CAM_TRANSFORM_SECONDS = 1.250f; // Save the camera transform every x seconds

	glm::mat4 modelBoat = glm::translate(glm::mat4(1.0f), boatPosition);
	glm::mat3 normalMatBoat = glm::mat3(glm::transpose(glm::inverse(modelBoat)));
	glm::mat4 modelIsland = glm::translate(glm::mat4(1.0f), islandPosition);
	glm::mat3 normalMatIsland = glm::mat3(glm::transpose(glm::inverse(modelIsland)));
	glm::mat4 modelMoon = glm::scale(glm::translate(glm::mat4(1.0f), moonFakePosition), glm::vec3(moonScale - 1.0f));
	
	model = glm::mat4(1.0f);
	translate = glm::translate(model, glm::vec3(0, waterlevel, 0));
	rotate = glm::rotate(model, glm::radians(90.0f), glm::vec3(-1, 0, 0));
	scale = glm::scale(model, glm::vec3(500.0f));
	glm::mat4 modelWaterPlane = translate * rotate * scale * model;
	glm::mat4 normalMatWaterPlane = glm::mat3(glm::transpose(glm::inverse(modelWaterPlane)));

	glm::mat4 modelText = glm::translate(glm::mat4(1.0f), glm::vec3(-0.1, -0.2, 0.0));

	float windCycle = 0.0f;
	startTime = glfwGetTime();
	while (!glfwWindowShouldClose(window))
	{

		//Calculating timing for waves and intro text fade in/fade out
		float currentTime = glfwGetTime();
		//if (firstTime) {
		//
		//	startTime = currentTime;
		//	SoundEngine->play2D(ambientSound, false);
		//	firstTime = 0;
		//}

		timeDifference = currentTime - lastTime;
		startDifference = std::fmin(currentTime - startTime, 5.0);
		lastTime = currentTime;
		//cameraSpeed = timeDifference * turbo;
		cameraSpeed = timeDifference * 1.75 * turbo;

		//FPS COUNTER
		//std::cout << 1.f / timeDifference << std::endl;

		if (!userControlledCamera)
			camera.updateAutoPathPosition(timeDifference);


		if (recordCameraMovement) {
			camTransformRecordingTime += timeDifference;
			if (camTransformRecordingTime > SAVE_CAM_TRANSFORM_SECONDS) {
				camTransformRecordingTime -= SAVE_CAM_TRANSFORM_SECONDS;
				cameraRecording.push_back(camera.getTransform());
			}
		}

		//in one second, waveFactor goes from 0 to 1000 in steps of 1 for each millisecond
		//we do the scaling to get more precise values (without scaling we would get 1 integer step per second, now we get 1000 integer steps per second)
		//we do the modulo so waveFactor does not go to infinity which could maybe lead to memory problems
		//a visible repetition of the pattern is only visible every 1000 seconds
		int time = (int)(currentTime * 1000);
		int waveFactor = time % 1000000;
	
		//similar to envelopes in music
		//1 second attack from 0 to 1, 3 seconds sustain on 1, 1 second release from 1 to 0
		//add x second delay to startTime so the text pops up x seconds later
		float attack = std::fmin(currentTime - startTime - 5.0, 1.0);
		float release = std::fmin(5.0 - (currentTime - startTime - 5.0), 1.0);
		float alphaFactor = std::fmax(attack + release - 1.0, 0.0);

		//play sound once after 1.5 seconds
		//if (firstSound && startDifference >= 4.5) {
		//	SoundEngine->play2D(areaSound, false);
		//	firstSound = 0;
		//}

		processInput(window);
		userInputUniforms.updateUserInputUB(currentUserInput);

		if (moonDisabled)
			lights = lightsWithoutMoon;
		else
			lights = lightsWithMoon;
		lightUniforms.updateLightUniforms(lights);


		windCycle += timeDifference;
		if (windCycle > 2 * 3.141592653589)
			windCycle = windCycle - 2 * 3.141592653589;

		//RENDER SHADOWMAPS FOR ALL LIGHTSOURCES_______________________________________
		depthFB.bind();
		glEnable(GL_DEPTH_TEST);
		glClear(GL_DEPTH_BUFFER_BIT);
		//set Uniforms
		depthMapShader.use();
		depthMapShader.setInt("numLayers", depthMaps.numLayers);
		
		for (int light = 0; light < lights.size(); light++) {
			depthMapShader.setFloatVec3("lightPositions[" + std::to_string(light) + "]", glm::vec3(lights[light].position));
			depthMapShader.setFloat("farPlanes[" + std::to_string(light) + "]", lights[light].far);
			for (int face = 0; face < 6; face++) {
				depthMapShader.setFloatMatrix4x4("lightSpaceMatrices[" + std::to_string(face + light * 6) + "]", lights[light].lightSpaceMatrices[face]);
			}
		}

		//island
		depthMapShader.setFloatMatrix4x4("model", modelIsland);
		depthMapShader.setFloatMatrix3x3("normalMat", normalMatIsland);
		island.Draw(depthMapShader);

		//boat
		depthMapShader.setFloatMatrix4x4("model", modelBoat);
		depthMapShader.setFloatMatrix3x3("normalMat", normalMatBoat);
		depthMapShader.setFloat("windCycle", windCycle);
		depthMapShader.setBool("windEnable", currentUserInput.windAct_tessEnable.x);
		
		boat.setTesselation(currentUserInput.windAct_tessEnable.y == 1);
		boat.drawNonTesselated(depthMapShader);

		if (currentUserInput.windAct_tessEnable.y) {
			depthTessShader.use();
			depthTessShader.setInt("numLayers", depthMaps.numLayers);
			for (int light = 0; light < lights.size(); light++) {
				depthTessShader.setFloatVec3("lightPositions[" + std::to_string(light) + "]", glm::vec3(lights[light].position));
				depthTessShader.setFloat("farPlanes[" + std::to_string(light) + "]", lights[light].far);
				for (int face = 0; face < 6; face++) {
					depthTessShader.setFloatMatrix4x4("lightSpaceMatrices[" + std::to_string(face + light * 6) + "]", lights[light].lightSpaceMatrices[face]);
				}
			}

			depthTessShader.setFloatMatrix4x4("model", modelBoat);
			depthTessShader.setFloatMatrix3x3("normalMat", normalMatBoat);
			depthTessShader.setFloat("windCycle", windCycle);
			depthTessShader.setBool("windEnable", currentUserInput.windAct_tessEnable.x);

			depthTessShader.setFloatMatrix4x4("modelMatrix", modelBoat);
			depthTessShader.setFloatMatrix4x4("viewMatrix", view);

			boat.drawTesselated(depthTessShader);
		}
		boat.setTesselation(false);

		depthFB.unbind();
		//render
		//_______________________________________RENDER SHADOWMAPS FOR ALL LIGHTSOURCES





		//RENDER WATER TEXTURES_______________________________________
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_CLIP_DISTANCE0);
		//REFLECTION___
		waterReflectionFB.bind();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glm::vec4 waterClipPlane(0, 1, 0, -waterlevel);

		//adjust camera
		camera.position = glm::vec3(camera.position.x, camera.position.y - (2 * (camera.position.y - waterlevel)), camera.position.z);
		camera.invertPitch();
		camera.calculateCameraVectors();
		view = camera.lookAt();
		projection = glm::perspective(glm::radians(camera.fov), (float)SCR_WIDTH / (float)SCR_HEIGHT, nearPlaneCamera, farPlaneCamera);
		viewProjectionClipUniforms.updateViewProjectionClipUniforms(view, projection, waterClipPlane);

		glm::mat4 viewProjection = projection * view;

		glActiveTexture(GL_TEXTURE15);
		glBindTexture(GL_TEXTURE_CUBE_MAP_ARRAY, depthMaps.ID);
		//volFog.compute(7,
		//	glm::inverse(viewProjection),
		//	glm::vec4(camera.position, 1.0),
		//	anisotropy_density,
		//	glm::vec4(0.002f, nearPlaneCamera, farPlaneCamera, 0.0f),
		//	15
		//);

		//RENDER SKYBOX
		skyboxShader.use();
		skyboxShader.setInt("FB_WIDTH", ENVIRONMENT_MAP_DIMENSIONS);
		skyboxShader.setInt("FB_HEIGHT", ENVIRONMENT_MAP_DIMENSIONS);
		//volFog.bindGridTexture(7);
		//skyboxShader.setInt("volFogVoxelGrid", volFog.getLastBoundIdx());
		skyboxShader.setBool("volFogEnable", false);
		skybox.render(skyboxShader);

		//RENDER LIGHT SOURCES
		/*
		lightSourceShader.use();
		for (Light light : lights) {
			model = glm::mat4(1.0f);
			glm::mat4 scale = glm::scale(model, glm::vec3(moonScale));
			glm::mat4 translate = glm::translate(model, glm::vec3(light.position));
			lightSourceShader.setFloatVec3("lightColor", light.diffuse);
			lightSourceShader.setFloatMatrix4x4("model", translate * scale * model);
			lightSourceShader.setFloatMatrix4x4("viewProjection", viewProjection);
			moon.Draw(lightSourceShader);
		}
		*/

		//RENDER LIGHTED OBJECTS
		lightingShader.use();
		//lightingShader.setInt("volFogVoxelGrid", volFog.getLastBoundIdx());  // ToDo: Check this texture index!
		lightingShader.setBool("volFogEnable", false);
		lightingShader.setBool("moonDisabled", moonDisabled);
		lightingShader.setFloatVec3("cameraPos", camera.position);
		lightingShader.setFloat("far_plane_light", lights[0].far);
		lightingShader.setFloat("far_plane_camera", farPlaneCamera);
		lightingShader.setFloat("near_plane_camera", nearPlaneCamera);
		glActiveTexture(GL_TEXTURE15);
		glBindTexture(GL_TEXTURE_CUBE_MAP_ARRAY, depthMaps.ID);
		lightingShader.setInt("depthMaps", 15);

		//boat
		lightingShader.setFloatMatrix4x4("model", modelBoat);
		lightingShader.setFloatMatrix3x3("normalMat", normalMatBoat);
		lightingShader.setFloatMatrix4x4("viewProjection", viewProjection);
		boat.Draw(lightingShader);

		//island
		glEnable(GL_CULL_FACE);
		lightingShader.setFloatMatrix4x4("model", modelIsland);
		lightingShader.setFloatMatrix3x3("normalMat", normalMatIsland);
		lightingShader.setFloatMatrix4x4("viewProjection", viewProjection);
		island.Draw(lightingShader);
		glDisable(GL_CULL_FACE);

		//readjust camera
		camera.position = glm::vec3(camera.position.x, camera.position.y - (2 * (camera.position.y - waterlevel)), camera.position.z);
		camera.invertPitch();
		camera.calculateCameraVectors();
		view = camera.lookAt();
		projection = glm::perspective(glm::radians(camera.fov), (float)SCR_WIDTH / (float)SCR_HEIGHT, nearPlaneCamera, farPlaneCamera);
		viewProjection = projection * view;
		//___REFLECTION


		//REFRACTION___
		
		waterRefractionFB.bind();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		waterClipPlane = glm::vec4(0, -1, 0, waterlevel);
		//update uniforms
		viewProjectionClipUniforms.updateViewProjectionClipUniforms(view, projection, waterClipPlane);

		volFog.compute(7,
			glm::inverse(projection * view),
			glm::vec4(camera.position, 1.0),
			anisotropy_density,
			glm::vec4(0.002f, nearPlaneCamera, farPlaneCamera, 0.0f),
			15,
			moonDisabled
		);

		//RENDER SKYBOX
		skyboxShader.use();
		skyboxShader.setInt("FB_WIDTH", ENVIRONMENT_MAP_DIMENSIONS);
		skyboxShader.setInt("FB_HEIGHT", ENVIRONMENT_MAP_DIMENSIONS);
		skyboxShader.setInt("volFogVoxelGrid", volFog.getLastBoundIdx());
		skyboxShader.setBool("volFogEnable", false);
		skybox.render(skyboxShader);

		
		//RENDER LIGHT SOURCES
		lightSourceShader.use();
		lightSourceShader.setInt("volFogVoxelGrid", volFog.getLastBoundIdx());
		lightSourceShader.setFloat("far_plane_camera", farPlaneCamera);
		lightSourceShader.setFloat("near_plane_camera", nearPlaneCamera);
		lightSourceShader.setFloatVec3("lightColor", lights[0].diffuse);
		lightSourceShader.setFloatMatrix4x4("model", modelMoon);
		lightSourceShader.setFloatMatrix4x4("viewProjection", viewProjection);
		moon.Draw(lightSourceShader);
		

		//RENDER LIGHTED OBJECTS
		lightingShader.use();
		lightingShader.setInt("volFogVoxelGrid", volFog.getLastBoundIdx());  // ToDo: Check this texture index!
		lightingShader.setBool("volFogEnable", false);
		lightingShader.setBool("moonDisabled", moonDisabled);
		lightingShader.setFloatVec3("cameraPos", camera.position);
		lightingShader.setFloat("far_plane_light", lights[0].far);
		lightingShader.setFloat("far_plane_camera", farPlaneCamera);
		lightingShader.setFloat("near_plane_camera", nearPlaneCamera);
		glActiveTexture(GL_TEXTURE15);
		glBindTexture(GL_TEXTURE_CUBE_MAP_ARRAY, depthMaps.ID);
		lightingShader.setInt("depthMaps", 15);


		//boat
		lightingShader.setFloatMatrix4x4("model", modelBoat);
		lightingShader.setFloatMatrix3x3("normalMat", normalMatBoat);
		lightingShader.setFloatMatrix4x4("viewProjection", viewProjection);
		boat.Draw(lightingShader);


		//island
		glEnable(GL_CULL_FACE);
		lightingShader.setFloatMatrix4x4("model", modelIsland);
		lightingShader.setFloatMatrix3x3("normalMat", normalMatIsland);
		lightingShader.setFloatMatrix4x4("viewProjection", viewProjection);
		island.Draw(lightingShader);
		glDisable(GL_CULL_FACE);
		//___REFRACTION
		//_______________________________________RENDER WATER TEXTURES










		//RENDER SCENE INTO HDR Framebuffer_______________________________________
		glEnable(GL_DEPTH_TEST);
		glDisable(GL_CLIP_DISTANCE0);
		fbHDR.bind();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		//RENDER SKYBOX
		skyboxShader.use();
		skyboxShader.setInt("volFogVoxelGrid", volFog.getLastBoundIdx());
		skyboxShader.setBool("volFogEnable", true);
		skyboxShader.setInt("FB_WIDTH", SCR_WIDTH);
		skyboxShader.setInt("FB_HEIGHT", SCR_HEIGHT);
		skybox.render(skyboxShader);

		//RENDER LIGHT SOURCES
		lightSourceShader.use();
		lightSourceShader.setInt("volFogVoxelGrid", volFog.getLastBoundIdx());
		lightSourceShader.setFloat("far_plane_camera", farPlaneCamera);
		lightSourceShader.setFloat("near_plane_camera", nearPlaneCamera);
		lightSourceShader.setFloatVec3("lightColor", lights[0].diffuse);
		lightSourceShader.setFloatMatrix4x4("model", modelMoon);
		lightSourceShader.setFloatMatrix4x4("viewProjection", viewProjection);
		moon.Draw(lightSourceShader);

		//RENDER LIGHTED OBJECTS
		//island
		lightingShader.use();
		lightingShader.setFloatMatrix4x4("model", modelIsland);
		lightingShader.setFloatMatrix3x3("normalMat", normalMatIsland);
		lightingShader.setFloatMatrix4x4("viewProjection", viewProjection);
		lightingShader.setBool("volFogEnable", true);
		lightingShader.setBool("moonDisabled", moonDisabled);
		island.Draw(lightingShader);

		//boat
		boat.setTesselation(currentUserInput.windAct_tessEnable.y == 1);
		modelWindShader.use();
		volFog.bindGridTexture(7);
		modelWindShader.setInt("volFogVoxelGrid", volFog.getLastBoundIdx());
		modelWindShader.setBool("volFogEnable", true);
		modelWindShader.setBool("moonDisabled", moonDisabled);
		modelWindShader.setFloatVec3("cameraPos", camera.position);
		modelWindShader.setFloat("far_plane_light", lights[0].far);
		modelWindShader.setFloat("far_plane_camera", farPlaneCamera);
		modelWindShader.setFloat("near_plane_camera", nearPlaneCamera);
		glActiveTexture(GL_TEXTURE15);
		glBindTexture(GL_TEXTURE_CUBE_MAP_ARRAY, depthMaps.ID);
		modelWindShader.setInt("depthMaps", 15);

		modelWindShader.setFloatMatrix4x4("model", modelBoat);
		modelWindShader.setFloatMatrix3x3("normalMat", normalMatIsland);

		modelWindShader.setFloat("windCycle", windCycle);
		modelWindShader.setBool("windEnable", currentUserInput.windAct_tessEnable.x);

		modelWindShader.setBool("tesselationEnabled", currentUserInput.windAct_tessEnable.y == 1);
		boat.drawNonTesselated(modelWindShader);

		if (currentUserInput.windAct_tessEnable.y) {
			modelTessWindShader.use();
			volFog.bindGridTexture(7);
			modelTessWindShader.setInt("volFogVoxelGrid", volFog.getLastBoundIdx());
			modelTessWindShader.setBool("volFogEnable", true);
			modelTessWindShader.setBool("moonDisabled", moonDisabled);
			modelTessWindShader.setFloatVec3("cameraPos", camera.position);
			modelTessWindShader.setFloat("far_plane_light", lights[0].far);
			modelTessWindShader.setFloat("far_plane_camera", farPlaneCamera);
			modelTessWindShader.setFloat("near_plane_camera", nearPlaneCamera);
			glActiveTexture(GL_TEXTURE15);
			glBindTexture(GL_TEXTURE_CUBE_MAP_ARRAY, depthMaps.ID);
			modelTessWindShader.setInt("depthMaps", 15);

			modelTessWindShader.setFloatMatrix4x4("model", modelBoat);
			modelTessWindShader.setFloatMatrix3x3("normalMat", normalMatBoat);

			modelTessWindShader.setFloat("windCycle", windCycle);
			modelTessWindShader.setBool("windEnable", currentUserInput.windAct_tessEnable.x);
			modelTessWindShader.setFloatMatrix4x4("modelMatrix", model);
			modelTessWindShader.setFloatMatrix4x4("viewMatrix", view);

			modelTessWindShader.setBool("tesselationEnabled", currentUserInput.windAct_tessEnable.y == 1);
			boat.drawTesselated(modelTessWindShader);
		}
		boat.setTesselation(false);



		
		//water
		waterShader.use();
		volFog.bindGridTexture(7);
		waterShader.setInt("volFogVoxelGrid", volFog.getLastBoundIdx());
		waterShader.setFloat("far_plane_camera", farPlaneCamera);
		waterShader.setFloat("near_plane_camera", nearPlaneCamera);
		waterShader.setFloatMatrix4x4("model", modelWaterPlane);
		waterShader.setFloatMatrix3x3("normalMat", normalMatWaterPlane);
		waterShader.setFloatVec3("cameraPos", camera.position);
		waterShader.setFloat("waveFactor", waveFactor);
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, waterReflectionFB.colorTex.ID);
		waterShader.setInt("reflectionTexture", 0);
		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, waterRefractionFB.colorTex.ID);
		waterShader.setInt("refractionTexture", 1);
		glActiveTexture(GL_TEXTURE2);
		glBindTexture(GL_TEXTURE_2D, waterDuDv.ID);
		waterShader.setInt("dudv", 2);
		glActiveTexture(GL_TEXTURE3);
		glBindTexture(GL_TEXTURE_2D, waterNormal.ID);
		waterShader.setInt("normal", 3);
		glActiveTexture(0);
		quad.Draw(waterShader);
		
		//_______________________________________RENDER SCENE INTO HDR Framebuffer





		//Render Scene on Quads_______________________________________
		glDisable(GL_DEPTH_TEST);
		glBindFramebuffer(GL_FRAMEBUFFER, 0);
		glViewport(0, 0, SCR_WIDTH, SCR_HEIGHT);
		glClear(GL_COLOR_BUFFER_BIT);

		//whole scene
		hdrShader.use();
		glActiveTexture(GL_TEXTURE0);
		fbHDR.colorTex.bind();
		hdrShader.setInt("quadTexture", 0);
		hdrShader.setFloat("exposure", exposure_gamma.x);
		hdrShader.setFloat("gamma", exposure_gamma.y);
		model = glm::mat4(1.0f);
		hdrShader.setFloatMatrix4x4("model", model);
		quad.Draw(hdrShader);
		fbHDR.colorTex.unbind();

		//textoverlay
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		textShader.use();
		areaName.bind();
		textShader.setFloatMatrix4x4("model", modelText);
		textShader.setInt("quadTexture", 0);
		textShader.setFloat("alphaFactor", alphaFactor);
		quad.Draw(textShader);
		glDisable(GL_BLEND);

		/*
		//first texturemonitor
		quadShader.use();
		glBindTexture(GL_TEXTURE_2D, waterReflectionFB.colorTex.ID);
		model = glm::mat4(1.0f);
		scale = glm::scale(model, glm::vec3(0.2));
		translate = glm::translate(model, glm::vec3(-0.8, 0.8, 0));
		model = translate * scale * model;
		quadShader.setFloatMatrix4x4("model", model);
		quadShader.setInt("quadTexture", 0);
		quad.Draw(quadShader);
		//second texturemonitor
		glBindTexture(GL_TEXTURE_2D, waterRefractionFB.colorTex.ID);
		model = glm::mat4(1.0f);
		scale = glm::scale(model, glm::vec3(0.2));
		translate = glm::translate(model, glm::vec3(-0.4, 0.8, 0));
		model = translate * scale * model;
		quadShader.setFloatMatrix4x4("model", model);
		quadShader.setInt("quadTexture", 0);
		quad.Draw(quadShader);
		*/
		//_______________________________________Render Scene on Quads

		glfwPollEvents();
		glfwSwapBuffers(window);
	}
	if (cameraRecording.size() != 0) {
		ofstream camRec("camRec.txt", ios_base::out | ios_base::trunc);
		camRec << SAVE_CAM_TRANSFORM_SECONDS << '\n';
		for (int i = 0; i < cameraRecording.size(); i++) {
			camRec << cameraRecording[i].position.x << ' ';
			camRec << cameraRecording[i].position.y << ' ';
			camRec << cameraRecording[i].position.z << ' ';
			camRec << cameraRecording[i].yaw << ' ';
			camRec << cameraRecording[i].pitch << '\n';
		}
		camRec.close();
	}

	glfwTerminate();
	return 0;
}

void processInput(GLFWwindow* window)
{
	if (glfwGetKey(window, GLFW_KEY_ESCAPE) == GLFW_PRESS)
		glfwSetWindowShouldClose(window, true);

	if (userControlledCamera) {
		if (glfwGetKey(window, GLFW_KEY_S) == GLFW_PRESS) {
			camera.moveBackwards(cameraSpeed);
		}
		if (glfwGetKey(window, GLFW_KEY_W) == GLFW_PRESS) {
			camera.moveForwards(cameraSpeed);
		}
		if (glfwGetKey(window, GLFW_KEY_A) == GLFW_PRESS) {
			camera.moveLeft(cameraSpeed);
		}
		if (glfwGetKey(window, GLFW_KEY_D) == GLFW_PRESS) {
			camera.moveRight(cameraSpeed);
		}
	}
	if (glfwGetKey(window, GLFW_KEY_LEFT_SHIFT) == GLFW_PRESS) {
		turbo = 15;
	}
	if (glfwGetKey(window, GLFW_KEY_LEFT_CONTROL) == GLFW_PRESS) {
		turbo = 0.4;
	}
	if (glfwGetKey(window, GLFW_KEY_LEFT_CONTROL) == GLFW_RELEASE && glfwGetKey(window, GLFW_KEY_LEFT_SHIFT) == GLFW_RELEASE) {
		turbo = 1;
	}
		
	if (glfwGetKey(window, GLFW_KEY_B) == GLFW_PRESS && !bPressed) {
		currentUserInput.volFogAct_blueNoise_filterAct_moveLight.y = currentUserInput.volFogAct_blueNoise_filterAct_moveLight.y == 1 ? 0 : 1;
		bPressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_B) == GLFW_RELEASE && bPressed) {
		bPressed = false;
	}
	
	if (glfwGetKey(window, GLFW_KEY_C) == GLFW_PRESS && !cPressed) {
		userControlledCamera = !userControlledCamera;
		if (!userControlledCamera)
			camera.startAutoPathing();
		cPressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_C) == GLFW_RELEASE && cPressed) {
		cPressed = false;
	}
	
	
	if (glfwGetKey(window, GLFW_KEY_F) == GLFW_PRESS && !fPressed) {
		currentUserInput.volFogAct_blueNoise_filterAct_moveLight.z = currentUserInput.volFogAct_blueNoise_filterAct_moveLight.z == 1 ? 0 : 1;
		fPressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_F) == GLFW_RELEASE && fPressed) {
		fPressed = false;
	}

	if (glfwGetKey(window, GLFW_KEY_O) == GLFW_PRESS && !oPressed) {
		cameraRecording = std::vector<CameraTransform>();
		recordCameraMovement = !recordCameraMovement;
		oPressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_O) == GLFW_RELEASE && oPressed) {
		oPressed = false;
	}

	if (glfwGetKey(window, GLFW_KEY_1) == GLFW_PRESS && !k1Pressed) {
		exposure_gamma.x = exposure_gamma.x - 0.1f;
		printf("Exposure: %f\n", exposure_gamma.x);
		k1Pressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_1) == GLFW_RELEASE && k1Pressed) {
		k1Pressed = false;
	}

	if (glfwGetKey(window, GLFW_KEY_2) == GLFW_PRESS && !k2Pressed) {
		exposure_gamma.x = exposure_gamma.x + 0.1f;
		printf("Exposure: %f\n", exposure_gamma.x);
		k2Pressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_2) == GLFW_RELEASE && k2Pressed) {
		k2Pressed = false;
	}

	if (glfwGetKey(window, GLFW_KEY_3) == GLFW_PRESS && !k3Pressed) {
		exposure_gamma.y = exposure_gamma.y - 0.1f;
		printf("Gamma: %f\n", exposure_gamma.y);
		k3Pressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_3) == GLFW_RELEASE && k3Pressed) {
		k3Pressed = false;
	}

	if (glfwGetKey(window, GLFW_KEY_4) == GLFW_PRESS && !k4Pressed) {
		exposure_gamma.y = exposure_gamma.y + 0.1f;
		printf("Gamma: %f\n", exposure_gamma.y);
		k4Pressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_4) == GLFW_RELEASE && k4Pressed) {
		k4Pressed = false;
	}

	/* Reset the camera position and restart the auto pathing */
	if (glfwGetKey(window, GLFW_KEY_R) == GLFW_PRESS && !rPressed) {
		camera.position = initialCameraPosition;
		camera.yaw = -90.0f;
		camera.pitch = 0.0f;
		camera.calculateCameraVectors();
		camera.startAutoPathing();
		userControlledCamera = false;
		rPressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_R) == GLFW_RELEASE && rPressed) {
		rPressed = false;
	}

	if (glfwGetKey(window, GLFW_KEY_T) == GLFW_PRESS && !tPressed) {
		currentUserInput.windAct_tessEnable.y = currentUserInput.windAct_tessEnable.y == 1 ? 0 : 1;
		tPressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_T) == GLFW_RELEASE && tPressed) {
		tPressed = false;
	}

	if (glfwGetKey(window, GLFW_KEY_U) == GLFW_PRESS && !uPressed) {
		currentUserInput.windAct_tessEnable.x = currentUserInput.windAct_tessEnable.x == 1 ? 0 : 1;
		uPressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_U) == GLFW_RELEASE && uPressed) {
		uPressed = false;
	}

	if (glfwGetKey(window, GLFW_KEY_V) == GLFW_PRESS && !vPressed) {
		currentUserInput.volFogAct_blueNoise_filterAct_moveLight.x = currentUserInput.volFogAct_blueNoise_filterAct_moveLight.x == 1 ? 0 : 1;
		vPressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_V) == GLFW_RELEASE && vPressed) {
		vPressed = false;
	}

	if (glfwGetKey(window, GLFW_KEY_M) == GLFW_PRESS && !mPressed) {
		moonDisabled = !moonDisabled;
		mPressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_M) == GLFW_RELEASE && mPressed) {
		mPressed = false;
	}

	if (glfwGetKey(window, GLFW_KEY_KP_ADD) == GLFW_PRESS && !plusPressed) {
		anisotropy_density.y = anisotropy_density.y + 1.0f;
		printf("Density: %f\n", anisotropy_density.y);
		plusPressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_KP_ADD) == GLFW_RELEASE && plusPressed) {
		plusPressed = false;
	}


	if (glfwGetKey(window, GLFW_KEY_KP_SUBTRACT) == GLFW_PRESS && !minusPressed) {
		if (anisotropy_density.y > 1.0f)
			anisotropy_density.y = anisotropy_density.y - 1.0f;
		printf("Density: %f\n", anisotropy_density.y);
		minusPressed = true;
	}

	if (glfwGetKey(window, GLFW_KEY_KP_SUBTRACT) == GLFW_RELEASE && minusPressed) {
		minusPressed = false;
	}

	if (glfwGetKey(window, GLFW_KEY_KP_9) == GLFW_PRESS && !kp9Pressed) {
		if (anisotropy_density.x <= 0.9f)
			anisotropy_density.x = anisotropy_density.x + 0.1f;
		printf("Anisotropy: %f\n", anisotropy_density.x);
		kp9Pressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_KP_9) == GLFW_RELEASE && kp9Pressed) {
		kp9Pressed = false;
	}


	if (glfwGetKey(window, GLFW_KEY_KP_6) == GLFW_PRESS && !kp6Pressed) {
		if (anisotropy_density.x > 0.1f)
			anisotropy_density.x = anisotropy_density.x - 0.1f;
		printf("Anisotropy: %f\n", anisotropy_density.x);
		kp6Pressed = true;
	}
	if (glfwGetKey(window, GLFW_KEY_KP_6) == GLFW_RELEASE && kp6Pressed) {
		kp6Pressed = false;
	}
	

	
}

void framebuffer_size_callback(GLFWwindow* window, int width, int height)
{
	glViewport(0, 0, width, height);
}

void mouse_callback(GLFWwindow* window, double xpos, double ypos) {

	if (firstMouse) {
		xpos = 0;
		ypos = 0;
		firstMouse = false;
	}

	xpos /= 10.0f;
	ypos /= 10.0f;
	if (userControlledCamera) {
		camera.yaw += xpos;
		camera.pitch -= ypos;

		if (camera.pitch > 89.0f)
			camera.pitch = 89.0f;
		if (camera.pitch < -89.0f)
			camera.pitch = -89.0f;

		camera.calculateCameraVectors();

		//std::cout << "xpos: " << xpos << " ypos: " << ypos << std::endl;
		//std::cout << "yaw: " << camera.yaw << " pitch: " << camera.pitch << std::endl;
		glfwSetCursorPos(window, 0, 0);
	}
}

void scroll_callback(GLFWwindow* window, double xoffset, double yoffset) {
	camera.fov -= (float)yoffset;
	if (camera.fov < 20.0f)
		camera.fov = 20.0f;
	if (camera.fov > 150.0f)
		camera.fov = 150.0f;

}