#pragma once

#include "shader.h"

struct VoxelGridDimension {
	uint32_t x, y, z;
};

class VolumetricFog {

public:
	VolumetricFog(const char * raymarchShPath, const char * lightInjectShPath, VoxelGridDimension voxelGridSize) {
		this->_raymarchSh = new Shader(raymarchShPath);
		this->_lightInjectSh = new Shader(lightInjectShPath);
		this->_voxelGridSize = voxelGridSize;

		createVoxelGrid();
		createBlueNoiseTexture();
	}

	/* Execute the volumetric fog calculation. Fills the voxel grid with data so it can be used for shading.
	
	activeTextureOffset: Set to free texture unit (if multiple are in use simultaneously)
	*/
	void compute(uint32_t activeTextureOffset,
		glm::mat4 invViewProj,
		glm::vec4 cameraPosition,
		glm::vec4 anisotropyDensity,
		glm::vec4 biasNearFar,
		int lightDepthMaps,
		bool moonDisabled
	) {
		lightInjection(activeTextureOffset, invViewProj, cameraPosition, anisotropyDensity, biasNearFar, lightDepthMaps, moonDisabled);
		raymarch(activeTextureOffset);
	}

	unsigned int getVoxelGridTexId() {
		return _voxelGridTexture;
	}

	void bindGridTexture(int textureIdx) {
		glActiveTexture(GL_TEXTURE0 + textureIdx);
		glBindTexture(GL_TEXTURE_3D, this->_voxelGridTexture);
		_lastBoundIdx = textureIdx;
	}

	int getLastBoundIdx() {
		return _lastBoundIdx;
	}
private:
	Shader * _raymarchSh = NULL;
	Shader * _lightInjectSh = NULL;
	VoxelGridDimension _voxelGridSize = {};
	unsigned int _voxelGridTexture = 0;
	unsigned int _blueNoiseTexture = 0;
	unsigned int _lastBoundIdx = -1;

	/* Creates a 3D texture that is used as the voxel grid	*/
	void createVoxelGrid() {
		glGenTextures(1, &(this->_voxelGridTexture));
		glActiveTexture(GL_TEXTURE0); // Needed?
		
		glBindTexture(GL_TEXTURE_3D, this->_voxelGridTexture);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		glTexImage3D(GL_TEXTURE_3D, 0, GL_RGBA16F, this->_voxelGridSize.x, this->_voxelGridSize.y, this->_voxelGridSize.z, 0, GL_RGBA, GL_HALF_FLOAT, NULL);

		glBindTexture(GL_TEXTURE_3D, this->_voxelGridTexture);

		glBindImageTexture(0, this->_voxelGridTexture, 0, GL_FALSE, 0, GL_READ_WRITE, GL_RGBA16F);
	}

	void createBlueNoiseTexture() {
		Texture2D blueNoise;
		blueNoise.loadImage("textures/blue_noise/LDR_RGBA_0.png"); // ToDo
		this->_blueNoiseTexture = blueNoise.ID;
	}

	/* Executes the light injection compute shader */
	void lightInjection(uint32_t activeTextureOffset,
		glm::mat4 invViewProj,
		glm::vec4 cameraPosition,
		glm::vec4 anisotropyDensity,
		glm::vec4 biasNearFar,
		int lightDepthMaps,
		bool moonDisabled
	) {
		this->_lightInjectSh->use();
		this->_lightInjectSh->setFloatMatrix4x4("invViewProj", invViewProj);
		this->_lightInjectSh->setFloatVec4("cameraPosition", cameraPosition);
		this->_lightInjectSh->setFloatVec4("biasNearFar", biasNearFar);
		this->_lightInjectSh->setBool("moonDisabled", moonDisabled);

		this->_lightInjectSh->setInt("lightDepthMaps", lightDepthMaps);

		this->_lightInjectSh->setFloatVec4("anisotropyDensity", anisotropyDensity);

		bindGridTexture(activeTextureOffset);
		glBindImageTexture(0, this->_voxelGridTexture, 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_RGBA16F);

		glActiveTexture(GL_TEXTURE1 + activeTextureOffset);
		glBindTexture(GL_TEXTURE_2D, this->_blueNoiseTexture);
		this->_lightInjectSh->setInt("blueNoiseSampler", 1 + activeTextureOffset); // ToDo


		const uint32_t LOCAL_SIZE_X = 8;
		const uint32_t LOCAL_SIZE_Y = 8;
		const uint32_t LOCAL_SIZE_Z = 1;

		glDispatchCompute(	(unsigned int)ceil((float)this->_voxelGridSize.x / (float)LOCAL_SIZE_X),
							(unsigned int)ceil((float)this->_voxelGridSize.y / (float)LOCAL_SIZE_Y),
							(unsigned int)ceil((float)this->_voxelGridSize.z / (float)LOCAL_SIZE_Z));
		glMemoryBarrier(GL_SHADER_IMAGE_ACCESS_BARRIER_BIT);
		
	}

	/* Executes the raymarch compute shader */
	void raymarch(uint32_t activeTextureOffset) {
		this->_raymarchSh->use();
		this->_raymarchSh->setInt("voxelGridIn", activeTextureOffset); // ToDo
		this->_raymarchSh->setFloatVec4("biasNearFar", glm::vec4(0.002f, 0.1f, 1000.0f, 1.0f)); // ToDo

		glActiveTexture(GL_TEXTURE0 + activeTextureOffset);
		glBindTexture(GL_TEXTURE_3D, this->_voxelGridTexture);
		glBindImageTexture(0, this->_voxelGridTexture, 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_RGBA16F);

		const uint32_t LOCAL_SIZE_X = 8;
		const uint32_t LOCAL_SIZE_Y = 8;
		const uint32_t LOCAL_SIZE_Z = 1;

		glDispatchCompute(	(unsigned int)ceil((float)this->_voxelGridSize.x / (float)LOCAL_SIZE_X),
							(unsigned int)ceil((float)this->_voxelGridSize.y / (float)LOCAL_SIZE_Y),
							1);
		glMemoryBarrier(GL_SHADER_IMAGE_ACCESS_BARRIER_BIT);
	}
};