#pragma once
#ifndef CAMERA_H
#define CAMERA_H

#include <sstream>
#include <string>

#include <glad/glad.h>
#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>

struct CameraTransform {
	glm::vec3 position;
	float yaw;
	float pitch;
};

class Camera {
public:
	glm::vec3 position;
	glm::vec3 worldUp;
	glm::vec3 center;//point that the camera looks at
	glm::vec3 front;
	glm::vec3 up;
	glm::vec3 right;
	float yaw; //left-right
	float pitch; // up-down
	float fov;

	Camera() {
		position = glm::vec3(0.0f,0.0f,5.0f);
		worldUp = glm::vec3(0.0f, 1.0f, 0.0f);
		front = glm::vec3(0.0f, 0.0f, -1.0f);
		right = glm::normalize(glm::cross(front, worldUp));
		up = glm::normalize(glm::cross(right, front));
		center = calculateCenter();
		yaw = -90.0f;
		pitch = 0.0f;
		fov = 45.0f;
	};

	glm::mat4 lookAt() {
		return glm::lookAt(position, center, up);
	};

	/* direction: -1 = left, +1 = right	*/
	void rotateYaw(float speed, int direction) {
		yaw += speed * direction;
		calculateCameraVectors();
	}

	/* direction: -1 = down, +1 = up	*/
	void rotatePitch(float speed, int direction) {
		pitch += speed * direction;
		calculateCameraVectors();
	}

	void moveForwards(float speed) {
		position = position + speed*front;
		calculateCenter();
	};

	void moveBackwards(float speed) {
		position = position - speed*front;
		calculateCenter();
	}

	void moveRight(float speed) {
		position = position + speed*right;
		calculateCenter();
	}

	void moveLeft(float speed) {
		position = position - speed*right;
		calculateCenter();
	}
	void invertPitch() {
		pitch = -pitch;
	}

	void calculateCameraVectors() {
		glm::vec3 direction;
		direction.x = cos(glm::radians(yaw)) * cos(glm::radians(pitch));
		direction.y = sin(glm::radians(pitch));
		direction.z = sin(glm::radians(yaw)) * cos(glm::radians(pitch));
		front = glm::normalize(direction);
		right = glm::normalize(glm::cross(front, worldUp));
		up = glm::normalize(glm::cross(right, front));
		center = calculateCenter();
	}

	/* Sets a new path for the auto pathing.
	* The fourth vector component can be used to set camera rotations without movements (set to w=1.0f).
	*/
	void setPath(std::vector<CameraTransform> path) {
		autoPath = path;
	}

	/* (Re)starts the auto pathing of the camera */
	void startAutoPathing() {
		if (autoPath.size() == 0)
			return;
		noPath = false;
		remainingPath = autoPath;
		currentPathGoal = remainingPath[0];
		lastPathGoal = this->getTransform();
		remainingPath.erase(remainingPath.begin());
	}


	void updateAutoPathPosition(float timeDelta) {
		if (noPath)
			return;

		accTimeDeltaAutoPath += timeDelta;
		float t = accTimeDeltaAutoPath / CAMERA_AUTO_PATH_TICKS;
		while (t >= 1.0f) {
			setNextPathGoal();
			t -= 1.0f;
			accTimeDeltaAutoPath = t * CAMERA_AUTO_PATH_TICKS;
		}

		this->position = lastPathGoal.position + (currentPathGoal.position - lastPathGoal.position) * t;
		this->yaw = lastPathGoal.yaw + (currentPathGoal.yaw - lastPathGoal.yaw) * t;
		this->pitch = lastPathGoal.pitch + (currentPathGoal.pitch - lastPathGoal.pitch) * t;
		calculateCameraVectors();
	}

	CameraTransform getTransform() {
		return CameraTransform{
				this->position,
				this->yaw,
				this->pitch };
	}

	void readCameraPathFile(std::string filePath) {
		std::ifstream camRec(filePath, std::ios_base::in);

		this->autoPath = std::vector<CameraTransform>();
		float x, y, z, yaw, pitch;
		std::string line;
		std::getline(camRec, line);
		CAMERA_AUTO_PATH_TICKS = std::stof(line);
		while (camRec >> x >> y >> z >> yaw >> pitch) {
			autoPath.push_back(CameraTransform{
				glm::vec3(x, y, z),
				yaw,
				pitch });
		}
		camRec.close();
	}


private:
	glm::vec3 calculateCenter() {
		center = position + front;
		return center;
	}


	/* For auto pathing */
	float CAMERA_AUTO_PATH_TICKS = 0.050f; // Tick length in seconds between auto pathing sub goals
	float accTimeDeltaAutoPath = 0.0f; // Accumulated time when auto pathing to interpolate between goals, goal is reached when accTimeDeltaAutoPath = CAMERA_AUTO_PATH_TICKS

	bool noPath = true; // Is set to true if current path is finished or startAutoPathing() hasn't been called, so camera won't move
	std::vector<CameraTransform> autoPath; // Path that the camera will follow
	std::vector<CameraTransform> remainingPath; // Remaining path of /autoPath/ which will be adapted after each check point is reached
	CameraTransform lastPathGoal; // Last checkpoint where the camera started pathing from
	CameraTransform currentPathGoal; // Current check point where the camera will walk towards

	void setNextPathGoal() {
		if (remainingPath.size() == 0) {
			noPath = true;
		}
		else {
			lastPathGoal = currentPathGoal;
			currentPathGoal = remainingPath[0];
			remainingPath.erase(remainingPath.begin());
		}
	}
};
#endif
