#pragma once

#include <GLAD/glad.h>

#include "texture.h"

/**
* This class encapsulates a Framebuffer with a color and depth attachement.
* The framebuffer has a texture bound for each attachment by default. The color attachment is of Format GL_RGB (thus no HDR).*
*/
class Framebuffer {
public:
	unsigned int ID;
    int width, height;
    Texture2D colorTex;
    Texture2D depthTex;

	Framebuffer(unsigned int width, unsigned int height) : colorTex(), depthTex() {

        this->width = width;
        this->height = height;

        glGenFramebuffers(1, &ID);
        glBindFramebuffer(GL_FRAMEBUFFER, ID);

        glBindTexture(GL_TEXTURE_2D, colorTex.ID);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, colorTex.ID, 0);
        glBindTexture(GL_TEXTURE_2D, 0);


        glBindTexture(GL_TEXTURE_2D, depthTex.ID);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, width, height, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, depthTex.ID, 0);
        glBindTexture(GL_TEXTURE_2D, 0);


        if (glCheckFramebufferStatus(GL_FRAMEBUFFER) == GL_FRAMEBUFFER_COMPLETE) {
            std::cout << "Framebuffer: " << ID << " complete" << std::endl;
        }
        else {
            std::cout << "ERROR::FRAMEBUFFER Framebuffer " << ID << " is not complete." << std::endl;
        }
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
	}

    void bind() {
        glBindFramebuffer(GL_FRAMEBUFFER, ID);
        glViewport(0, 0, width, height);
    }
    void unbind() {
        glBindFramebuffer(GL_FRAMEBUFFER, 0);

    }
};
/**
* This class encapsulates a Framebuffer with a color and depth attachement.
* The framebuffer has a texture bound for each attachment by default.
* In contrast to "Framebuffer", "FramebufferHDR" has a color attachment of Format GL_RGB16F that allows for HDR (High dynamic range).
* Apart from that, it is the same as "Framebuffer".
*/
class FramebufferHDR {
public:
    unsigned int ID;
    int width, height;
    Texture2D colorTex;
    Texture2D depthTex;

    FramebufferHDR(unsigned int width, unsigned int height) : colorTex(), depthTex() {

        this->width = width;
        this->height = height;

        glGenFramebuffers(1, &ID);
        glBindFramebuffer(GL_FRAMEBUFFER, ID);

        glBindTexture(GL_TEXTURE_2D, colorTex.ID);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB16F, width, height, 0, GL_RGB, GL_FLOAT, NULL);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, colorTex.ID, 0);
        glBindTexture(GL_TEXTURE_2D, 0);


        glBindTexture(GL_TEXTURE_2D, depthTex.ID);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, width, height, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, depthTex.ID, 0);
        glBindTexture(GL_TEXTURE_2D, 0);


        if (glCheckFramebufferStatus(GL_FRAMEBUFFER) == GL_FRAMEBUFFER_COMPLETE) {
            std::cout << "Framebuffer: " << ID << " complete" << std::endl;
        }
        else {
            std::cout << "ERROR::FRAMEBUFFER Framebuffer " << ID << " is not complete." << std::endl;
        }
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
    }

    void bind() {
        glBindFramebuffer(GL_FRAMEBUFFER, ID);
        glViewport(0, 0, width, height);
    }
    void unbind() {
        glBindFramebuffer(GL_FRAMEBUFFER, 0);

    }
};