#pragma once

#include <GLAD/glad.h>

#include <glm/vec3.hpp>
#include <glm/vec4.hpp>

#include <vector>
#include <string>

#include "shader.h"



class Light {
public:
	std::string type;
	glm::vec4 position;

	glm::vec3 ambient;
	glm::vec3 diffuse;
	glm::vec3 specular;

	glm::vec3 attenuation;

	//these transformation matrices transform a point in world space to the "light space", which is essentially
	//a camera space from the lights point view
	//this is used for rendering the scene from the lights point of view for example when creating a shadow cubemap for the light
	std::vector<glm::mat4> lightSpaceMatrices;
	float near, far;


	Light(std::vector<glm::vec3> color, glm::vec4 position, glm::vec3 attenuation) {
		this->ambient = color[0];
		this->diffuse = color[1];
		this->specular = color[2];

		this->position = position;
		this->attenuation = attenuation;

		if (position[3] == 1)
			this->type = "point";
		else
			this->type = "directional";

		//the angle has to be 90 degrees so each face of the shadow cubemap  covers exactly 90 degrees
		//this way the edges of the faces line up correctly
		//the aspect ratio is set to 1.0 because we assume that the cubemap face textures are square
		this->near = 1.0f;
		this->far = 200.0f;
		glm::mat4 projectionMatrix = glm::perspective(glm::radians(90.0f), 1.0f, near, far);

		//directions for the camera to look  at when rendering sphere environment map
		//the order of the directions is right, left, top, bottom, back, front
		lightSpaceMatrices.push_back(projectionMatrix *
			glm::lookAt(glm::vec3(position), glm::vec3(position) + glm::vec3(1.0, 0.0, 0.0), glm::vec3(0.0, -1.0, 0.0)));
		lightSpaceMatrices.push_back(projectionMatrix *
			glm::lookAt(glm::vec3(position), glm::vec3(position) + glm::vec3(-1.0, 0.0, 0.0), glm::vec3(0.0, -1.0, 0.0)));
		lightSpaceMatrices.push_back(projectionMatrix *
			glm::lookAt(glm::vec3(position), glm::vec3(position) + glm::vec3(0.0, 1.0, 0.0), glm::vec3(0.0, 0.0, 1.0)));
		lightSpaceMatrices.push_back(projectionMatrix *
			glm::lookAt(glm::vec3(position), glm::vec3(position) + glm::vec3(0.0, -1.0, 0.0), glm::vec3(0.0, 0.0, -1.0)));
		lightSpaceMatrices.push_back(projectionMatrix *
			glm::lookAt(glm::vec3(position), glm::vec3(position) + glm::vec3(0.0, 0.0, 1.0), glm::vec3(0.0, -1.0, 0.0)));
		lightSpaceMatrices.push_back(projectionMatrix *
			glm::lookAt(glm::vec3(position), glm::vec3(position) + glm::vec3(0.0, 0.0, -1.0), glm::vec3(0.0, -1.0, 0.0)));


	};

	void setPosition(glm::vec4 pos) {
		this->position = pos;
		updateLightSpaceMatrices();
	};

	void updateLightSpaceMatrices() {
		//the angle has to be 90 degrees so each face of the shadow cubemap  covers exactly 90 degrees
		//this way the edges of the faces line up correctly
		//the aspect ratio is set to 1.0 because we assume that the cubemap face textures are square
		glm::mat4 projectionMatrix = glm::perspective(glm::radians(90.0f), 1.0f, near, far);

		//directions for the camera to look  at when rendering sphere environment map
		//the order of the directions is right, left, top, bottom, back, front
		lightSpaceMatrices[0] = (projectionMatrix *
			glm::lookAt(glm::vec3(position), glm::vec3(position) + glm::vec3(1.0, 0.0, 0.0), glm::vec3(0.0, -1.0, 0.0)));
		lightSpaceMatrices[1] = (projectionMatrix *
			glm::lookAt(glm::vec3(position), glm::vec3(position) + glm::vec3(-1.0, 0.0, 0.0), glm::vec3(0.0, -1.0, 0.0)));
		lightSpaceMatrices[2] = (projectionMatrix *
			glm::lookAt(glm::vec3(position), glm::vec3(position) + glm::vec3(0.0, 1.0, 0.0), glm::vec3(0.0, 0.0, 1.0)));
		lightSpaceMatrices[3] = (projectionMatrix *
			glm::lookAt(glm::vec3(position), glm::vec3(position) + glm::vec3(0.0, -1.0, 0.0), glm::vec3(0.0, 0.0, -1.0)));
		lightSpaceMatrices[4] = (projectionMatrix *
			glm::lookAt(glm::vec3(position), glm::vec3(position) + glm::vec3(0.0, 0.0, 1.0), glm::vec3(0.0, -1.0, 0.0)));
		lightSpaceMatrices[5] = (projectionMatrix *
			glm::lookAt(glm::vec3(position), glm::vec3(position) + glm::vec3(0.0, 0.0, -1.0), glm::vec3(0.0, -1.0, 0.0)));
	}
};