#pragma once

#ifndef SPHERE_H
#define SPHERE_H

#include <GLAD/glad.h>

#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>

#include <iostream>
#include <memory>
#include <array>
#include <vector>

#include "shader.h"
#include "texture.h"

/*
* An abstraction of a sphere
* The sphere is getting rendered on a billboard (square) with raytracing
* the raytracing uses the center and radius of the sphere to render the sphere on the quad
* this class is meant to be used with the "billboard" shader
*/
class Sphere
{
public:
    std::vector<float> vertices;
    // center coordinates in world space
    glm::vec3 center;
    float radius;

    Sphere(glm::vec3 center, float radius) {
        this->center = center;
        this->radius = radius;

        //Vertex Data of the Billboard (It is a square)
        vertices = {
            //Triangle 1
            -radius,  +radius, 0,
            -radius,  -radius, 0,
            +radius,  -radius, 0,

            //Triangle 2
            -radius,  +radius, 0,
            +radius,  -radius, 0,
           +radius,  +radius, 0
        };

        glGenVertexArrays(1, &VAO);
        glBindVertexArray(VAO);

        glGenBuffers(1, &VBO);
        glBindBuffer(GL_ARRAY_BUFFER, VBO);
        glBufferData(GL_ARRAY_BUFFER, vertices.size()*sizeof(float), vertices.data(), GL_STATIC_DRAW);

        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(float), 0);
        glEnableVertexAttribArray(0);

        glBindBuffer(GL_ARRAY_BUFFER, 0);
        glBindVertexArray(0);
    }

    /*
    * renders the sphere reflecting the given Cubemap
    */
    void render(Shader& shader,Cubemap& cubemap) {

        glBindVertexArray(VAO);
        shader.use();
        glActiveTexture(0);
        glBindTexture(GL_TEXTURE_CUBE_MAP, cubemap.ID);
        shader.setInt("skybox", 0);
        shader.setFloatVec3("center", center);
        shader.setFloat("radius", radius);

        glDrawArrays(GL_TRIANGLES, 0, vertices.size());

        glBindTexture(GL_TEXTURE_CUBE_MAP, 0);


    }
    /*
    * renders the sphere reflecting the given Environment map
    */
    void render(Shader& shader, EnvironmentMap& environment) {

        glBindVertexArray(VAO);
        shader.use();
        glActiveTexture(0);
        glBindTexture(GL_TEXTURE_CUBE_MAP, environment.ID);
        shader.setInt("skybox", 0);
        shader.setFloatVec3("center", center);
        shader.setFloat("radius", radius);

        glDrawArrays(GL_TRIANGLES, 0, vertices.size());

        glBindTexture(GL_TEXTURE_CUBE_MAP, 0);


    }
private:
    unsigned int VAO, VBO;
};

#endif // SPHERE_H
