#pragma once

#ifndef TEXTURE_H
#define TEXTURE_H

#include <GLAD/glad.h>

#include <iostream>
#include <string>
#include <filesystem>
#include <unordered_map>

//Class for loading a 2D Texture
class Texture2D
{
public:
	unsigned int ID;
	//generate Texture
	Texture2D(bool repeat = true) {
		glGenTextures(1, &ID);
		// set the texture wrapping/filtering options
		glBindTexture(GL_TEXTURE_2D, ID);
		if (repeat) {
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		}
		else {
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		}
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glBindTexture(GL_TEXTURE_2D, 0);
	}
	/*
	loads the image from the given path to the texture
	*/
	void loadImage(const char* srcPath) {

		glBindTexture(GL_TEXTURE_2D, ID);

		// load and generate the texture
		int width, height, nrChannels;
		stbi_set_flip_vertically_on_load(true);
		unsigned char* data = stbi_load(srcPath, &width, &height, &nrChannels, 0);
		if (data) {
			unsigned int imageFormat = (nrChannels == 4) ? GL_RGBA : GL_RGB;
			glTexImage2D(GL_TEXTURE_2D, 0, imageFormat, width, height, 0, imageFormat, GL_UNSIGNED_BYTE, data);
			glGenerateMipmap(GL_TEXTURE_2D);
			std::cout << "LOADED TEXTURE " << ID << std::endl;

		}
		else {
			std::cout << "ERROR::FAILED TO LOAD TEXTURE IMAGE" << std::endl;
		}
		stbi_image_free(data);

		glBindTexture(GL_TEXTURE_2D, 0);
	}

	void createEmptyTexture(unsigned int width, unsigned int height) {
		glBindTexture(GL_TEXTURE_2D, ID);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
		glBindTexture(GL_TEXTURE_2D, 0);

	}

	void bind() {
		glBindTexture(GL_TEXTURE_2D, ID);
	}
	void unbind() {
		glBindTexture(GL_TEXTURE_2D, 0);
	}
};

class Cubemap
{
public:
	unsigned int ID;
	  Cubemap() {
		  glGenTextures(1, &ID);
		  glBindTexture(GL_TEXTURE_CUBE_MAP, ID);
		  glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		  glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		  glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		  glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		  glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
		  glBindTexture(GL_TEXTURE_CUBE_MAP, 0);

	  }

	  /*
	  loads texture images from a directory into the cubemap faces
	  the files have to be named: right, left, top, bottom, back, front
	  (plus their file extension, so for example "right.jpg" or "right.png")
	  */
	  void loadImages(const char* srcDir) {

		  glBindTexture(GL_TEXTURE_CUBE_MAP, ID);

		  int width, height, nrChannels;
		  std::unordered_map<std::string, unsigned int> faces = {
					{"right",  GL_TEXTURE_CUBE_MAP_POSITIVE_X},
					{"left",   GL_TEXTURE_CUBE_MAP_NEGATIVE_X},
					{"top",    GL_TEXTURE_CUBE_MAP_POSITIVE_Y},
					{"bottom", GL_TEXTURE_CUBE_MAP_NEGATIVE_Y},
					{"back",   GL_TEXTURE_CUBE_MAP_POSITIVE_Z},
					{"front",  GL_TEXTURE_CUBE_MAP_NEGATIVE_Z}
		  };

		  if (std::filesystem::exists(srcDir) && std::filesystem::is_directory(srcDir)) {
			  std::filesystem::directory_iterator directory(srcDir);
			  for (const auto& file : directory) {
				  std::string filePath = file.path().string();
				  std::string fileName = file.path().filename().string();

				  std::string face = "";
				  for (char c : fileName) {
					  if (c == '.')
						  break;
					  face.append(1, c);
				  };
				  std::cout << "Loading Cubemap texture: " << filePath << std::endl;
				  unsigned char* data = stbi_load(filePath.c_str(), &width, &height, &nrChannels, 0);
				  if (data) {
					  unsigned int imageFormat = (nrChannels == 4) ? GL_RGBA : GL_RGB;
					  glTexImage2D(faces[face], 0, GL_RGB, width, height, 0, imageFormat, GL_UNSIGNED_BYTE, data);
					  stbi_image_free(data);
				  }
				  else {
					  std::cout << "FAILED TO LAOD CUBEMAP AT PATH: " << filePath << std::endl;
					  stbi_image_free(data);
				  }
			  }
		  }
		  else {
			  std::cout << "ERROR:TEXTURE::CUBEMAP the given path " << srcDir << " does not exist or is not a directory" << std::endl;
		  }
		  glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
  	  }

	  /*
	  * copies an existing 2D Texture into a given CubeMap face of the Cubemap instance 
	  */
	  void loadTexture(unsigned int cubemap_face, unsigned int texture_id) {

		  glBindTexture(GL_TEXTURE_CUBE_MAP, ID);
		  glBindTexture(GL_TEXTURE_2D, texture_id);

		  //read data from 2d texture
		  int width, height, internalFormat;
		  glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &width);
		  glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &height);
		  //glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_INTERNAL_FORMAT, &internalFormat);
		  internalFormat = GL_RGB;
		  std::vector<unsigned char> buffer(width * height * 3);
		  glGetTexImage(GL_TEXTURE_2D, 0, internalFormat, GL_UNSIGNED_BYTE, buffer.data());

		  //copy data to cubemap face
		  glTexImage2D(cubemap_face, 0, internalFormat, width, height, 0, internalFormat, GL_UNSIGNED_BYTE, buffer.data());

		  glBindTexture(GL_TEXTURE_2D, 0);
		  glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
	  }
};


class DepthCubemap
{
public:
	unsigned int ID;
	unsigned int dimensions;

	DepthCubemap(unsigned int dimensions) {
		this->dimensions = dimensions;

		glGenTextures(1, &ID);
		glBindTexture(GL_TEXTURE_CUBE_MAP, ID);

		//assign each face an empty depth texture
		for (unsigned int i = 0; i < 6; ++i)
			glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, GL_DEPTH_COMPONENT32,
				dimensions, dimensions, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);

		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

		glBindTexture(GL_TEXTURE_CUBE_MAP, 0);

	}
};

/*
* Encapsulates a Cubemap Array where each layer (i.e. Cubemap) holds a Cubemap with depth information
* this is used for rendering multiple depthMaps (that used for shadow mapping) in one render pass using
* a geometry shader
* is initialized with empty textures
*/
class DepthCubemap3D
{
public:
	unsigned int ID;
	unsigned int dimensions;
	unsigned int numLayers;

	DepthCubemap3D(unsigned int dimensions, unsigned int numLayers) {
		this->dimensions = dimensions;
		this->numLayers = numLayers;

		glGenTextures(1, &ID);
		glBindTexture(GL_TEXTURE_CUBE_MAP_ARRAY, ID);

		//assign each face an empty depth texture
		glTexImage3D(GL_TEXTURE_CUBE_MAP_ARRAY, 0, GL_DEPTH_COMPONENT, dimensions, dimensions, 6 * numLayers, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);

		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARRAY, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARRAY, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARRAY, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARRAY, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARRAY, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

		glBindTexture(GL_TEXTURE_CUBE_MAP_ARRAY, 0);

	}
};

class EnvironmentMap
{
public:
	unsigned int ID;
	unsigned int dimensions;

	EnvironmentMap(unsigned int dimensions) {
		this->dimensions = dimensions;

		glGenTextures(1, &ID);
		glBindTexture(GL_TEXTURE_CUBE_MAP, ID);
		for (unsigned int i = 0; i < 6; ++i) {
			glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, GL_RGB,
				dimensions, dimensions, 0, GL_RGB, GL_UNSIGNED_INT, NULL);
		}
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

		glBindTexture(GL_TEXTURE_CUBE_MAP, 0);


	}
};

/*
* Encapsulates a Cubemap Array where each layer (i.e. Cubemap) can hold a Cubemap with color information
* this is used for rendering multiple environment maps using a geometry shader
* is initialized with empty textures
*/
class EnvironmentMap3D
{
public:
	unsigned int ID;
	unsigned int dimensions;
	unsigned int numLayers;

	EnvironmentMap3D(unsigned int dimensions, unsigned int numLayers) {
		this->dimensions = dimensions;
		this->numLayers = numLayers;

		glGenTextures(1, &ID);
		glBindTexture(GL_TEXTURE_CUBE_MAP_ARRAY, ID);

		//assign each face an empty depth texture
		glTexImage3D(GL_TEXTURE_CUBE_MAP_ARRAY, 0, GL_RGB, dimensions, dimensions, 6 * numLayers, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);

		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARRAY, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARRAY, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARRAY, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARRAY, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARRAY, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

		glBindTexture(GL_TEXTURE_CUBE_MAP_ARRAY, 0);


	}
};

class WaterTextureArray {
public:
	unsigned int ID;
	unsigned int dimensions;
	unsigned int numLayers;

	WaterTextureArray(unsigned int dimensions) {
		this->dimensions = dimensions;
		this->numLayers = 2;

		glGenTextures(1, &ID);
		glBindTexture(GL_TEXTURE_2D_ARRAY, ID);

		for (int layer = 0; layer < numLayers; layer++) {
			glTexImage3D(GL_TEXTURE_2D_ARRAY, 0, GL_RGB, dimensions, dimensions, layer, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
		}
	}
};
#endif