#pragma once

#include <GLAD/glad.h>

#include <glm/glm.hpp>
#include <glm/vec3.hpp>
#include <glm/vec4.hpp>
#include <glm/gtc/type_ptr.hpp>

#include <vector>

#include "light.h"

/**
* Encapsulates Uniform Buffer Objects
*/
class UniformBuffer {
public:
	unsigned int ID;

	void bind() {
		glBindBuffer(GL_UNIFORM_BUFFER, ID);
	}
	void unbind() {
		glBindBuffer(GL_UNIFORM_BUFFER, 0);

	}
};

/**
* Uniform Buffer Object for the Light Uniforms
* The Buffer has space for 11 light objects where each light consists of:
* position, ambient color, diffuse color, specular color,  attenuation
* Additionally, the first variable stores the number of Lights. This is needed in the shaders for looping through the number of Lights correctly.
*/
class LightUniformBuffer : public UniformBuffer {
public:
	/*
	* creates the Uniforms Buffer Object and allocates the buffer space.
	*/
	LightUniformBuffer() {
		glGenBuffers(1, &ID);
		glBindBuffer(GL_UNIFORM_BUFFER, ID);
		glBufferData(GL_UNIFORM_BUFFER, 56 * sizeof(glm::vec4) + 11 * sizeof(float), NULL, GL_STATIC_DRAW);
		glBindBufferRange(GL_UNIFORM_BUFFER, 1, ID, 0, 56 * sizeof(glm::vec4) + 11 * sizeof(float));
		glBindBuffer(GL_UNIFORM_BUFFER, 0);
	}

	/**
	* fills the Buffer with the light data from the given Light vector
    * when calculating the byte offsets, keep in mind the std140 buffer memory layout
	* the offset always has to be a multiple of 16 bytes
	* therefore, even if the integer numLights only uses 4 bytes, position[0] will still have an offset of 16 bytes
	* also, a vec3 induces the same offset as vec4 (which is 16 bytes)
	*/
	void updateLightUniforms(vector<Light>& lights) {

		bind();
		int numLights = lights.size();
		glBufferSubData(GL_UNIFORM_BUFFER, 0, sizeof(int), &numLights);
		for (int i = 0; i < numLights; i++) {
			glBufferSubData(GL_UNIFORM_BUFFER, (i + 1) * sizeof(glm::vec4), sizeof(glm::vec4), glm::value_ptr(lights[i].position));
			glBufferSubData(GL_UNIFORM_BUFFER, (i + 12) * sizeof(glm::vec4), sizeof(glm::vec4), glm::value_ptr(lights[i].ambient));
			glBufferSubData(GL_UNIFORM_BUFFER, (i + 23) * sizeof(glm::vec4), sizeof(glm::vec4), glm::value_ptr(lights[i].diffuse));
			glBufferSubData(GL_UNIFORM_BUFFER, (i + 34) * sizeof(glm::vec4), sizeof(glm::vec4), glm::value_ptr(lights[i].specular));
			glBufferSubData(GL_UNIFORM_BUFFER, (i + 45) * sizeof(glm::vec4), sizeof(glm::vec4), glm::value_ptr(lights[i].attenuation));
			glBufferSubData(GL_UNIFORM_BUFFER, (i + 56) * sizeof(glm::vec4), sizeof(float), &lights[i].far);
		}
		unbind();
	}
};

/**
* Uniform Buffer Object for View Matrix, Projection Matrix and ClipPlane uniform variables
*/
class ViewProjectionClipUniformBuffer : public UniformBuffer {
public:
	/*
	* creates the Uniforms Buffer Object and allocates the buffer space.
	*/
	ViewProjectionClipUniformBuffer() {
		glGenBuffers(1, &ID);
		glBindBuffer(GL_UNIFORM_BUFFER, ID);
		glBufferData(GL_UNIFORM_BUFFER, sizeof(glm::vec4) + 2 * sizeof(glm::mat4), NULL, GL_STATIC_DRAW);
		glBindBufferRange(GL_UNIFORM_BUFFER, 0, ID, 0, sizeof(glm::vec4) + 2 * sizeof(glm::mat4));
		glBindBuffer(GL_UNIFORM_BUFFER, 0);
	}

	/*
	* sets view and projection matrix variables in buffer
	*/
	void updateViewProjectionUniforms(glm::mat4& view, glm::mat4& projection) {
		bind();
		glBufferSubData(GL_UNIFORM_BUFFER, 0, sizeof(glm::mat4), glm::value_ptr(view));
		glBufferSubData(GL_UNIFORM_BUFFER, sizeof(glm::mat4), sizeof(glm::mat4), glm::value_ptr(projection));
		unbind();
	}
	/*
	* sets view and projection matrix and ClipPlane variables in buffer
	*/
	void updateViewProjectionClipUniforms(glm::mat4& view, glm::mat4& projection, glm::vec4 clipPlane) {
		bind();
		glBufferSubData(GL_UNIFORM_BUFFER, 0, sizeof(glm::mat4), glm::value_ptr(view));
		glBufferSubData(GL_UNIFORM_BUFFER, sizeof(glm::mat4), sizeof(glm::mat4), glm::value_ptr(projection));
		glBufferSubData(GL_UNIFORM_BUFFER, 2 * sizeof(glm::mat4), sizeof(glm::vec4), glm::value_ptr(clipPlane));
		unbind();
	}
};
struct UserInput {
	glm::ivec4 volFogAct_blueNoise_filterAct_moveLight;
	glm::ivec4 windAct_tessEnable;
};
class UserInputUniformBuffer : public UniformBuffer {
	public:
		UserInputUniformBuffer() {
			glGenBuffers(1, &ID);
			glBindBuffer(GL_UNIFORM_BUFFER, ID);
			glBufferData(GL_UNIFORM_BUFFER, sizeof(glm::vec4), NULL, GL_STATIC_DRAW);
			glBindBufferRange(GL_UNIFORM_BUFFER, 2, ID, 0, sizeof(glm::vec4));
			glBindBuffer(GL_UNIFORM_BUFFER, 0);
		}
		
		void updateUserInputUB(UserInput ui) {
			bind();
			glBufferSubData(GL_UNIFORM_BUFFER, 0, sizeof(glm::ivec4), &ui.volFogAct_blueNoise_filterAct_moveLight);
			glBufferSubData(GL_UNIFORM_BUFFER, sizeof(glm::ivec4), sizeof(glm::ivec4), &ui.windAct_tessEnable);
			unbind();
		}
};
